import type { Completion } from '@codemirror/autocomplete';
import type { CompletionContext } from '@codemirror/autocomplete';
import { parser as dentakuParser } from './dentaku.grammar';
import { LanguageSupport } from '@codemirror/language';
import { linter } from '@codemirror/lint';
import type { LintSource } from '@codemirror/lint';
import { LRLanguage } from '@codemirror/language';
import { Options } from 'tsup';

export declare const builtInFunctions: readonly ["acos", "acosh", "asin", "asinh", "atan", "atan2", "atanh", "cbrt", "cos", "cosh", "erf", "erfc", "exp", "frexp", "gamma", "hypot", "ldexp", "lgamma", "log", "log10", "log2", "sin", "sinh", "sqrt", "tan", "tanh", "if", "and", "or", "xor", "not", "switch", "min", "max", "sum", "avg", "count", "round", "roundup", "rounddown", "abs", "case", "left", "right", "mid", "len", "substitute", "concat", "contains", "map", "filter", "all", "any", "pluck"];

export declare const builtInOperators: string[];

export declare const default_alias: Options | Options[] | ((overrideOptions: Options) => Options | Options[] | Promise<Options | Options[]>);

/**
 * The full Dentaku language support, including autocompletion and linting.
 *
 * @example
 * const codeMirrorExtensions = [
 *   ...otherExtensions,
 *   dentaku(),
 * ];
 */
export declare function dentaku({ linterOptions: { codeMirrorConfig, ...linterOptions }, completionOptions, }?: {
    linterOptions?: {
        codeMirrorConfig?: Parameters<typeof linter>[1];
    } & DentakuLinterOptions;
    completionOptions?: DentakuLanguageCompletionOptions;
}): LanguageSupport;

declare function dentakuCompletions({ variableEntries, makeEntryForBuiltInFunctions, makeEntryForBuiltInOperators, }?: DentakuLanguageCompletionOptions): (context: CompletionContext) => {
    options: (false | Completion | null)[];
    from: number;
    validFor?: undefined;
} | {
    options: Completion[];
    from: number;
    validFor: RegExp;
} | undefined;
export { dentakuCompletions }
export { dentakuCompletions as dentakuCompletions_alias_1 }

/**
 * Only the parser and syntax highlighter for Dentaku.
 *
 * Use this only if you want to omit/overwrite either linting or autocompletion.
 */
export declare const dentakuLanguage: LRLanguage;

declare interface DentakuLanguageCompletionOptions {
    /**
     * CodeMirror `Completion` objects for known variables.
     *
     * @example
     * ["a", "b"].map(name => ({ label: name, type: "variable" }))
     */
    variableEntries?: Array<Completion>;
    /**
     * Converter of built-in Dentaku functions into `Completion` objects.
     * Allows you to add additional information to completions like descriptions.
     */
    makeEntryForBuiltInFunctions?: (name: (typeof builtInFunctions)[number]) => Completion | null | false;
    /**
     * Converter of built-in Dentaku operators into `Completion` objects.
     * Allows you to add additional information to completions like descriptions.
     */
    makeEntryForBuiltInOperators?: (name: (typeof builtInOperators)[number]) => Completion | null | false;
}
export { DentakuLanguageCompletionOptions }
export { DentakuLanguageCompletionOptions as DentakuLanguageCompletionOptions_alias_1 }

declare function dentakuLinter({ knownVariables, messages, }?: DentakuLinterOptions): LintSource;
export { dentakuLinter }
export { dentakuLinter as dentakuLinter_alias_1 }

declare interface DentakuLinterOptions {
    /** List of variable names to not treat as undefined. */
    knownVariables?: Array<string>;
    /** Custom error messages. */
    messages?: ErrorMessages;
}
export { DentakuLinterOptions }
export { DentakuLinterOptions as DentakuLinterOptions_alias_1 }

export { dentakuParser }

/** Message codes that can be overridden with custom messages. */
declare interface ErrorMessages {
    /** Generic "invalid syntax" message when no better message can be inferred. */
    invalidSyntax: string;
    /** Missing a closing bracket in a list of function arguments. */
    closingBracketMissing: string;
    /** Expected a comma before this function argument. */
    expectedCommaBefore: string;
    /** Expected an operator before this expression. */
    expectedOperatorBefore: string;
    /** The function name needs parentheses to be called. */
    callParenthesesMissing: string;
    /** This variable is not defined. */
    undefinedVariable: string;
    /** Expected at least one parameter in a function call. */
    expectedAtLeastOneParameter: string;
    /** Expected that many parameters, found this many. */
    parameterCountMismatch: (expectedAmount: number, actualAmount: number) => string;
}
export { ErrorMessages }
export { ErrorMessages as ErrorMessages_alias_1 }

/** Special character reserved for invalid syntax trees. */
declare const invalidSyntax = "\u26A0";
export { invalidSyntax }
export { invalidSyntax as invalidSyntax_alias_1 }

export { }
